import { addTransitionState, removeTransitionState } from 'diffhtml';

var assign = Object.assign,
    keys = Object.keys;


var eventsToTransitionName = {
  attributechanged: 'attributeChanged',
  textchanged: 'textChanged'
};

// Store maps of elements to handlers that are associated to transitions.
var transitionsMap = {
  attached: new Map(),
  detached: new Map(),
  replaced: new Map(),
  attributechanged: new Map(),
  textchanged: new Map()
};

// Internal global transition state handlers, allows us to bind once and match.
var boundHandlers = [];

/**
 * Binds inline transitions to the parent element and triggers for any matching
 * nested children.
 */
export default function inlineTransitions() {
  var options = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : {};

  // Monitors whenever an element changes an attribute, if the attribute is a
  // valid state name, add this element into the related Set.
  var attributeChanged = function attributeChanged(domNode, name, oldVal, newVal) {
    var prefix = name.toLowerCase().slice(0, 2);

    // Don't bother with non-events.
    if (prefix !== 'on') {
      return;
    }

    // Normalize the event name to
    name = name.toLowerCase().slice(2);

    var map = transitionsMap[name];
    var isFunction = typeof newVal === 'function';

    // Abort early if not a valid transition or if the new value exists, but
    // isn't a function.
    if (!map || newVal && !isFunction) {
      return;
    }

    // Add or remove based on the value existence and type.
    map[isFunction ? 'set' : 'delete'](domNode, newVal);
  };

  var subscribe = function subscribe() {
    addTransitionState('attributeChanged', attributeChanged);

    // Add a transition for every type.
    keys(transitionsMap).forEach(function (name) {
      var map = transitionsMap[name];
      var transitionName = eventsToTransitionName[name] || name;

      var handler = function handler(child) {
        for (var _len = arguments.length, rest = Array(_len > 1 ? _len - 1 : 0), _key = 1; _key < _len; _key++) {
          rest[_key - 1] = arguments[_key];
        }

        // If there are no elements to match here, abort.
        if (!map.size) {
          return;
        }

        // If the child element triggered in the transition is the root
        // element, this is an easy lookup for the handler.
        if (map.has(child)) {
          return map.get(child).apply(undefined, [child, child].concat(rest));
        }
        // The last resort is looping through all the registered elements to
        // see if the child is contained within. If so, it aggregates all the
        // valid handlers and if they return Promises return them into a
        // `Promise.all`.
        else {
            var retVal = [];

            // Last resort check for child.
            map.forEach(function (fn, element) {
              if (element.contains(child)) {
                retVal.push(fn.apply(child, [element].concat(child, rest)));
              }
            });

            var hasPromise = retVal.some(function (ret) {
              return Boolean(ret && ret.then);
            });

            // This is the only time the return value matters.
            if (hasPromise) {
              return Promise.all(retVal);
            }
          }
      };

      // Save the handler for later unbinding.
      boundHandlers.push(handler);

      // Add the state handler.
      addTransitionState(transitionName, handler);
    });
  };

  // This will unbind any internally bound transition states.
  var unsubscribe = function unsubscribe() {
    // Unbind all the transition states.
    removeTransitionState('attributeChanged', attributeChanged);

    // Remove all elements from the internal cache.
    keys(transitionsMap).forEach(function (name) {
      var map = transitionsMap[name];
      var transitionName = eventsToTransitionName[name] || name;

      // Unbind the associated global handler.
      removeTransitionState(transitionName, boundHandlers.shift());

      // Empty the associated element set.
      map.clear();
    });

    // Empty the bound handlers.
    boundHandlers.length = 0;
  };

  return assign(function inlineTransitionsTask() {}, { subscribe: subscribe, unsubscribe: unsubscribe });
}